/*******************************************************************************

docbegin:


*  Name of source file:         j_ima_mosaic_params.c

*  Version of source file:      7.0.2

*  Parent component:            j_ima_mosaic_main

*  Programmer:                  Jerome Chenevez

*  Affiliation:                 Danish National Space Center

*  Contact address:             Juliane Maries Vej 30
*                               DK-2100 Copenhagen, Denmark
*                               Phone: +45 35325703  FAX: +45 35362475
*                               jerome@dnsc.dk

*  Purpose:                     Get the parameters and open files for the 
*                               j_ima_mosaic executable

*  Origin date:                 20041206

*  Update history:              5.0    20041206 
*				5.0.1  20050419  Now also combine iros input 
*						 images in cnts/cm2/s
*				5.0.2  20050603  Urgent fixes for OSA 5 by SP 
*                                                SPRs: 4174 (part.), 4176 (part.), 4190, 4191, 4205
*                               5.0.3  20050620  Fix SPRs 4174, 4175 4176, 4222
*                               5.0.4  20050622  Fix SPR 4244
*                               5.1.0  20051014  SCREW 1772 : base the image selection on VARIANCE maps
*                               6.0.0  20060329  Fix SPR 4377 : resolve energy selection problem
*						 SCREW 1850 : Use correct WCS routines allowing large mosaics
*                               6.0.1  20060908  Fix SPR 4301 : new parameter dolBPL
*                               6.0.2  20060915  SCREW 1918 : Copy j_ima_iros IMATYPE keywords to output
*                               6.0.3  20060927  SCREW 1929 : Gather all energy bands and mosaic products in one single output file
*                               6.0.4  20061114  Partial fix of SPR 4603 by forcing correct output image units (cnts/dm2/s)
*                               6.0.5  20061206  SPR 4603 is now fixed
*                                                SPR 4625 : Some selection strings defined too short
*                                                SPR 4632 : Allow variable number of input images by dynamic memory allocation
*                               7.0.0  20070707  SCREW 2008 Allow mixing of JMX1 and JMX2 mosaics
*						 SCREW 2024 Optionnal printing of input ScW list in the HEADERs of mosaic file
*                               7.0.1  20070820  SPR 4711: Allow empty dolBPL when combining Jem-X mosaics
*                               7.0.2  20070824  SPR 4714: Initialisations previous to DALallocateDataBuffer
   
*  Type of element:             Function

*  Part of package:             j_ima_mosaic

*  Name:                        j_ima_mosaic_params

*  Calls:

*  Returns:

*  Required components:         PIL, RIL, CFITSIO

docend:

Various remarks:

Flow of program:

Error codes:

****************************************************************/

#include "j_ima_mosaic.h"

int j_ima_mosaic_params(
                   char *groupDOL,
                   char *imSelect,
                   char *dolBPL,
                   char *mosaic,
                   double *radius,
                   double *emin,
                   double *emax,
                   double *diamout,
                   double *cdelt, 
                   double *centerA,
                   double *centerD,
                   int *plot_T, 
                   int *plot_I,
                   int *plot_V,
                   int *plot_S,
                   int *liScW,
                   int *comos,
               /*  int *clobber,  */
                   int *chatter,
                   int status)
{
/* Immediate return if called while executable is in error condition */
  if(status != ISDC_OK) return(status);

/* Read input parameters */
/*  status = PILInit(argc, argv);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Error_2, "%d: PIL initialisation error", status);
    return status;
  }
*/
  status = PILGetString("inObsGrp", groupDOL);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read inObsGrp parameter");
    return (status);
  }
  status = PILGetString("mapSelect", imSelect);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read intensity map parameter");
    return (status);
  }
  status = PILGetString("dolBPL", dolBPL);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read vignetting map parameter");
    return (status);
  }
  status = PILGetReal("radiusSelect", radius);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read radiuSelect parameter");
    return (status);
  }
  status = PILGetReal("eminSelect", emin);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read eminSelect parameter");
    return (status);
  }
  status = PILGetReal("emaxSelect", emax);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read emaxSelect parameter");
    return (status);
  }
  status = PILGetReal("diameter", diamout);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read diameter parameter");
    return (status);
  }
  status = PILGetReal("cdelt", cdelt);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read cdelt parameter");
    return (status);
  }
  status = PILGetReal("RAcenter", centerA);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read RAcenter parameter");
    return (status);
  }
  status = PILGetReal("DECcenter", centerD);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read DECcenter parameter");
    return (status);
  }
  status = PILGetString("outfile", mosaic);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read outfile parameter");
    return (status);
  }
  status = PILGetBool("viewTime", plot_T);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read viewTime parameter");
    return (status);
  }
  status = PILGetBool("viewIntens", plot_I);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read viewIntens parameter");
    return (status);
  }
  status = PILGetBool("viewVar", plot_V);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read viewVar parameter");
    return (status);
  }
  status = PILGetBool("viewSig", plot_S);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read viewSig parameter");
    return (status);
  }
  status = PILGetBool("print_ScWs", liScW);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read print_ScWs parameter");
    return (status);
  }
  status = PILGetBool("moscomb", comos);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read moscomb parameter");
    return (status);
  }
  status = PILGetInt("chatter", chatter);
  if (status != ISDC_OK) {
    RILlogMessage(NULL, Warning_2, "Could not read chatter parameter");
    return (status);
  }

/* PILClose(status); */

  return ( status );
}

/*****************************************************************

docbegin:


*  Name of source file:         j_ima_mosaic_work.c

*  Version of source file:      7.0.1

*  Parent component:            j_ima_mosaic_main

*  Programmer:                  Jerome Chenevez

*  Affiliation:                 Danish National Space Center

*  Contact address:             Juliane Maries Vej 30
*                               DK-2100 Copenhagen, Denmark
*                               Phone: +45 35325703  FAX: +45 35362475
*                               jerome@dnsc.dk

*  Purpose:                     Read the input images and produce 
*                               mosaic images

*  Origin date:                 20050131

*  Update history:              5.0    20041206   
*				5.0.1  20050419  Now also combine iros input 
*						 images in cnts/cm2/s
*				5.0.2  20050603  Urgent fixes for OSA 5 by SP 
*                                                SPRs: 4174 (part.), 4176 (part.), 4190, 4191, 4205
*                               5.0.3  20050620  Fix SPRs 4174, 4175 4176, 4222
*                               5.0.4  20050622  Fix SPR 4244
*                               5.1.0  20051014  SCREW 1772 : base the image selection on VARIANCE maps
*                               6.0.0  20060329  Fix SPR 4377 : resolve energy selection problem
*						 SCREW 1850 : Use correct WCS routines allowing large mosaics
*                               6.0.1  20060908  Fix SPR 4301 : new parameter dolBPL
*                               6.0.2  20060915  SCREW 1918 : Copy j_ima_iros IMATYPE keywords to output
*                               6.0.3  20060927  SCREW 1929 : Gather all energy bands and mosaic products in one single output file
*                               6.0.4  20061114  Partial fix of SPR 4603 by forcing correct output image units (cnts/dm2/s)
*                               6.0.5  20061206  SPR 4603 is now fixed
*                                                SPR 4625 : Some selection strings defined too short
*                                                SPR 4632 : Allow variable number of input images by dynamic memory allocation
*                               7.0.0  20070707  SCREW 2008 Allow mixing of JMX1 and JMX2 mosaics
*						 SCREW 2024 Optionnal printing of input ScW list in the HEADERs of mosaic file
*                               7.0.1  20070820  SPR 4711: Allow empty dolBPL when combining Jem-X mosaics
*                               7.0.2  20070824  SPR 4714: Initialisations previous to DALallocateDataBuffer
   
*  Type of element:             Function

*  Part of package:             j_ima_mosaic

*  Name:                        j_ima_mosaic_work

*  Calls:

*  Returns:

*  Required components:         PIL, RIL, CFITSIO

docend:

Various remarks:

Flow of program:

Error codes:

****************************************************************/

#include "j_ima_mosaic.h"

int j_ima_mosaic_work(
                   char *groupDOL,
                   char *imSelect,
                   char *dolBPL,
                   char *mosaic,
                   double radius,
                   double emin,
                   double emax,
                   double diamout,
                   double cdelt, 
                   double centerA,
                   double centerD,
                   int plot_T, 
                   int plot_I,
                   int plot_V,
                   int plot_S,
                   int liScW,
                   int comos,
               /*  int clobber,  */
                   int chatter,
                   int status)
{
/*  int liScW=1;*/
  int i=0, j=0, k=0, loop=0, overloop=0, cnt=0, cntImages=0;
  int *loopM=NULL, *rank=NULL, testM=0;
  int xdim=XD, ydim=YD;
  int dimx=(int)(sqrt(MOSAICSIZE)), dimy=(int)(sqrt(MOSAICSIZE)), dimos=(int)(sqrt(MOSAICSIZE));
  int numAxes=2, nbAxes=3, slice=0;
  int p=0, q=0, pix[2]={0, 0};
  int index=0, index1=0, index2=0;
  int status1=ISDC_OK;
  int isub, nsub, psub=0, qsub=0;
  int numImages=0, numINT=0, numVAR=0, numE=0, numEXP=0;
  int nex=1, numHDU=0, numIHDU=0, numVHDU=0, numEHDU=0, numTHDU=0;
  int num_E=0, e_dol=0;
  int useA=0;
  unsigned short *vign3D=NULL;
  long vigncode = 0L, *vslice=NULL;
  long axes[3] = {(long)(XD), (long)(YD), (long)(NB_E)};
  long startValues[2] = {1L, 1L}, endValues[2] = {(long)(XD), (long)(YD)};
  long numValues=0L;
  long vignStart[3] = {1L, 1L, 1L};
  long vignEnd[3] = {(long)(XD), (long)(YD), (long)(NB_E)}, nbValues=0L;
  long endMosaic[2] = {(long)(XD), (long)(YD)};
  float d_a=0., d_b=0., d_c=0., d_D=0.;
  float wei=1., *r_wei=NULL, *rwei=NULL, *rweiMax=NULL;
  float pixfac=1., reduc=1.; 
  float dsub=0., dp=0., dq=0.;
  float cdelta = 0.016667, cdeltb = 0.06667;
  double maxvign=0.;
  double pi=3.1415, dr=1., roll=0.;
  double ra_o=0., de_o=0., ra_x=0., de_x=0.;
  double t1[3][3], t2[3][3], t[3][3];
  double t1_o[3][3], t2_o[3][3], t_o[3][3], o_t[3][3];
  double NaN = NAN;
  double diamosaic=15.;
  double *exp_t=NULL;
  double *crpix1=NULL, *crpix2=NULL;
  double *cd1_1=NULL, *cd2_1=NULL,*cd1_2=NULL, *cd2_2=NULL;
  double *crval1=NULL, *crval2=NULL;
  double crval1_min=0., crval1_max=0., crval2_min=0., crval2_max=0.;
  double crval1_o=0., crval2_o=0.;
  double crpix1_o=0., crpix2_o=0.;
  double cd1_1_o=0., cd1_2_o=0., cd2_1_o=0., cd2_2_o=0.;
  double yz[2], ayz[2], vi[3], ve[3], vi_o[3];
  double *min_e=NULL, *max_e=NULL, prec_e=0.001;
  double **e_min=NULL, *e_minT=NULL, **e_max=NULL, *e_maxT=NULL;
  double **vign=NULL, *vignT=NULL, **vigntmp=NULL, *vigntmpT=NULL;
  double **pval_i=NULL, *pval_iT=NULL, **var_i=NULL, *var_iT=NULL;
  double **exp_v=NULL, *exp_vT=NULL, *expos=NULL;
  double **exp_r=NULL, *exp_rT=NULL, *time=NULL;
  double *pvalue=NULL, *variance=NULL;
  char extSelect[256], addSelect[256];
  char intSelect[256], varSelect[256], nrjSelect[256]; 
  char expSelect[256], timSelect[256]; 
  char fitstring[256];
  char vigndol[256];
  char imaType[19];
  char imaCase[9], basic_str[]="BASIC_01", fine_str[]="FINE_01";
  char iros_str[]="IROS001";
  char imaDescr[9];
  char raw_str[]="RAWINT", rec_str[]="RECTIF", src_str[]="RECONS";
  char template[]="JMXi-MOSA-IMA-IDX.tpl";
  char **swid=NULL, *swidT=NULL, **swid_red=NULL, *swid_redT=NULL, **scw=NULL, *scwT=NULL;
  char instName[10];
  char nameBPL[256], *BPLptr=NULL;
  char *imName=NULL;
  void *arrayPtr=NULL;
  void *arrayBuff=NULL;

  dal_dataType DALtype = DAL_DOUBLE;
  dal_dataType DALdble = DAL_DOUBLE;
  dal_element *dolPtr = NULL;
  dal_element *vigPtr = NULL;
  dal_element *intPtr = NULL;
  dal_element *nrjPtr = NULL;
  dal_element *expinp = NULL;
  dal_element *skyPtr = NULL;
  dal_element *pvlPtr = NULL;
  dal_element *varPtr = NULL;
  dal_element *expPtr = NULL;
  dal_element *timPtr = NULL;
  dal_element *vmoPtr = NULL;
  dal_element *mapPtr = NULL;
  dal_element *texPtr = NULL;
  dal_element *extPtr = NULL;
  dal_element *imList = NULL;
  dal_element *imAdrs = NULL;
  dal_element *idxSel = NULL;
  dal_element *intSel = NULL;
  dal_element *varSel = NULL;
  dal_element *nrjSel = NULL;
  dal_element *expSel = NULL;
  dal_element *timSel = NULL;
  dal_element *group = NULL;

  LogFile fileRef = NULL;

  Instrument instrument = INST_NOT_RECOGNIZED;

/* Necessary for WCS routines */
  int offscl=0;
  double cd_i[4], cd_o[4];
  double cdelt_i=0.025, rot=0.;
  double Xpos=0.,Ypos=0., Xpix=0.,Ypix=0.;
  char ctype_1[]="RA---TAN", ctype_2[]="DEC--TAN";
  struct WorldCoor *wcs_inp, *wcs_out;

/* STATEMENTS BEGIN */

  if (!comos) {BPLptr = strrchr(dolBPL, '/' ); strcpy(nameBPL, BPLptr+1);}

/* LOOK FOR NUMBER AND SIZE OF IMAGES FROM THE OBSERVATION GROUP */
  do{
    status=DALallocateDataBuffer((void**)&imName,20*sizeof(char),status); if (ISDC_OK!=status) break;

    status = DAL3GENopenISDCGroup(groupDOL, &group, status); /* Open obs group to get Ptr */
    if (status != ISDC_OK){
      RILlogMessage(NULL, Error_3, "Cannot open observation group %s!  status= %d", groupDOL, status);
      return status;
    }

    if (!comos){
      status = DALattributeGet(group,"INSTRUME",DAL_CHAR, (void*)instName,NULL,NULL,status);
      if (ISDC_OK!=status){
        RILlogMessage(NULL,Error_3,"Cannot read INSTRUME(nt)!");
        return status;
      }

      instrument = instrumentGetCode(instName);
  
      switch (instrument) {
    	  case JMX1:
              strcpy(imName,"JMX1-SKY.-IMA");
  	      strcpy(template,"JMX1-MOSA-IMA-IDX.tpl");
  	      break;
    
  	  case JMX2:
              strcpy(imName,"JMX2-SKY.-IMA");
  	      strcpy(template,"JMX2-MOSA-IMA-IDX.tpl");
  	      break;  
  
  	  case JEMX:
              strcpy(imName,"JEMX-SKY.-IMA");
  	      strcpy(template,"JEMX-MOSA-IMA-IDX.tpl");
  	      break;
  
  	  case INST_NOT_RECOGNIZED:
  	       RILlogMessage(NULL,Error_3,"INSTRUME(nt) %s not recognized!", instName);
  	       status=J_MOS_BAD_INSTRUMENT;
  	       break;
  	     
  	  default:
	       RILlogMessage(NULL,Error_1, "INSTRUME(nt) %s is not supported!", instName);
	       status=J_MOS_BAD_INSTRUMENT;
	       break;
      }
      if (ISDC_OK!=status) break;
    }
    else{
      strcpy(imName,"JMX?-MOSA-IMA");
      strcpy(template,"JMX1-MOSA-IMA-IDX.tpl");
    }

    status = DAL3GENindexFindMember(group, imName, "LAST", &numImages, &imList, status); /* List last extension of input files to get their number and a Ptr */
    if (status==DAL3GEN_INDEX_DOESNT_MATCH){
      RILlogMessage(NULL, Warning_1, "No data of type %s found in the input group!", imName);
      break;  
    } else if (ISDC_OK!=status) {
      RILlogMessage(NULL, Error_3, "Error %d determining number of input images!", status);
      break;
    }
    if (1>numImages) {
      RILlogMessage(NULL, Warning_2, "No data of type %s matched selection!", imName);
      break;
    }
    if (comos) RILlogMessage(NULL, Log_1, "Found %d input mosaic files with data of type %s", numImages, imName);
    else RILlogMessage(NULL, Log_1, "Found %d ScWs with data of type %s", numImages, imName);
  
    /* SP: Different treatment if there's only one image */
    if (numImages>1) status = DAL3GENindexGetMember(imList, NULL, 1, &imAdrs, status); /* Select 1st member of the listed input files to get a ptr */
    else imAdrs=imList;

    if(status != ISDC_OK){
      RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get 1st extension! status: %d", status);
      return( status );
    }
  
    status = DALattributeGetInt(imAdrs, "NAXIS1", &numValues, NULL, NULL, status); /* Get input images dimensions */
    if(status != ISDC_OK){
      RILlogMessage(NULL, Warning_2,"Cannot get attribute NAXIS1... status: %d", status);
      return status;
    }
    else RILlogMessage(NULL, Log_1, "Input images dimensions are: %ld * %ld pixels", numValues, numValues);
  
    status = DALattributeGetChar(imAdrs, "IMATYPE", imaType, NULL, NULL, status); /* Check the 1st skymap IMATYPE in imList*/
    if (status != ISDC_OK){
      RILlogMessage(NULL, Warning_2, "*** IMATYPE not found! Stop with status: %d ***", status);
      return status;                       /* Stop the program */
    }

  if (!comos){

/************************************************ THIS PART IS OBSOLETE (V 6.0.2) *****************************************
    if (strncmp(imaType, "INTENSITY", 6) == 0){
      status = DALattributeGetChar(imAdrs, "IMADESCR", imaDescr, NULL, NULL, status);
      if ( (status != ISDC_OK) || ( (strncmp(imaDescr, raw_str, 7) != 0) && (strncmp(imaDescr, rec_str, 7) != 0) ) ){
        RILlogMessage(NULL, Warning_2, "*** IMADESCR for INTENSITY image not found! Stop with status: %d ***", status);
        return status; 
      }
    }
    else if (strncmp(imaType, "RECONSTRUCTED", 6) == 0){
      status = DALattributeGetChar(imAdrs, "IMADESCR", imaDescr, NULL, NULL, status);
      if ( (status != ISDC_OK) || (strncmp(imaDescr, src_str, 5) != 0) ){
        RILlogMessage(NULL, Warning_2, "*** IMADESCR for RECONSTRUCTED image not found! Stop with status: %d ***", status);
        return status; 
      }
    } else sprintf(imaDescr, "%s", rec_str);

************************************************ THIS PART IS OPTIONAL ***************************************************/
    status = DALattributeGetInt(imAdrs, "VIGNCOR", &vigncode, NULL, NULL, status); /* Check if VIGNCOR is present */
    if(status != ISDC_OK){
      if (chatter > 4) RILlogMessage(NULL, Warning_1, "*** VIGNCOR not found -> set to 0! ***");
      vigncode = 0L;
      status = ISDC_OK;
    }
    if ( (int)(vigncode) == 1) RILlogMessage(NULL, Log_2, "Assume vignetting corrections already applied on input Intensity and Variance maps");
    if ( plot_T && (int)(vigncode) == 1) RILlogMessage(NULL, Log_2, "... will apply the same for the effective exposure mosaic map");

/**************************************************************************************************************************/

    status = DALattributeGetChar(imAdrs, "IMACASE", imaCase, NULL, NULL, status); /* Check the IMACASE */
    if ( (strncmp(imaCase, basic_str, 4) == 0) || (strncmp(imaCase, fine_str, 4) == 0) ){
      if (strncmp(imSelect, src_str, 5) == 0) RILlogMessage(NULL, Warning_1, "The IMACASE and the parameter mapSelect do not match eachother!");
      status = DALattributeGetReal(imAdrs, "ROLL", &roll, NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2,"Cannot get attribute ROLL! Stop with status: %d", status);
        return status;                     /* Stop the program */
      }
      useA = -1;
    }
    else if ( (strncmp(imaCase, iros_str, 4) == 0) ){
      if ( strncmp(imSelect, raw_str, 5)==0 || strncmp(imSelect, "RAW_INT", 5)==0 ) useA = 0;         /*** The options are for ***/
      else if ( strncmp(imSelect, rec_str, 5)==0 || strncmp(imSelect, "RAW_REC", 5)==0 ) useA = 1;    /*  backward compatibilty  */
      else if ( strncmp(imSelect, src_str, 5)==0 || strncmp(imSelect, "RES+SRC", 5)==0 ) useA = 2;    /***         only        ***/
      else strcpy(imSelect, imaType);
      if (plot_T){
        status = DALattributeGetChar(imAdrs, "VIGNDOL", vigndol, NULL, NULL, status); /* Eventually check VIGNDOL */
        if(status != ISDC_OK){
          RILlogMessage(NULL, Warning_1, "*** VIGNDOL not found ***");
          status = ISDC_OK;
        }
        else if (strcmp(vigndol, nameBPL) != 0){
          RILlogMessage(NULL, Warning_1, "The vignetting info indicated by the parameter dolBPL is: %s", nameBPL);
          RILlogMessage(NULL, Warning_1, "The vignetting info indicated by the keyword VIGNDOL is : %s", vigndol);
          RILlogMessage(NULL, Warning_2, "*** dolBPL and VIGNDOL are not exactly identical... ***");
          RILlogMessage(NULL, Warning_2, "Going to use dolBPL - This will only affect the effective exposure map");
        }
        strcpy(vigndol, dolBPL);
      }
    }
    else {
      RILlogMessage(NULL, Warning_2, "*** IMACASE not recognized! status: %d ***", status);
      return status;
    }

    RILlogMessage(NULL, Log_1, "Selecting %s Intensity sky maps of type: %s", imaCase, imSelect);

   } else{
      if ( strncmp(imSelect, raw_str, 5)==0 || strncmp(imSelect, "RAW_INT", 5)==0 ) useA = 0;
      else if ( strncmp(imSelect, rec_str, 5)==0 || strncmp(imSelect, "RAW_REC", 5)==0 ) useA = 1;
      else if ( strncmp(imSelect, src_str, 5)==0 || strncmp(imSelect, "RES+SRC", 5)==0 ) useA = 2;
      else strcpy(imSelect, imaType);
   } /* endof if !comos */

  }while(0);

/* INITIALISATIONS */
  pi = acos (-1.);
  dr = pi/180.;
  isub = 2;
  nsub = 2*isub + 1;
  dsub = 1./nsub;
  reduc = dsub*dsub;

  if (comos){
    liScW=0;
    vigncode = 0L;
  }

  /* V 6.0.5: Starting memory allocations of variable size arrays */
  if ( DALallocateDataBuffer((void**)&exp_t, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&crpix1, numImages*sizeof(double),status)!= ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&crpix2, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&cd1_1, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&cd2_1, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&cd1_2, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&cd2_2, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&crval1, numImages*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&crval2, numImages*sizeof(double),status) != ISDC_OK ) return status;

  ydim = xdim = (int)(numValues);   /* Assuming that the input skymaps always have square dimensions */
  endValues[0] = numValues;
  endValues[1] = numValues;

  for (i=0; i<numImages; i++){
    crpix1[i] = crpix2[i] = 0;
    exp_t[i] = 0.;
    cd1_1[i] = cd2_1[i] = cd1_2[i] = cd2_2[i] = 0.;
    crval1[i] = crval2[i] = 0.;
  }
  for (i=0; i<MOSAICSIZE; i++){
    pval_o[i] = 0.;
    var_o[i] = 0.;
    t_exp[i] = 0.;
    v_exp[i] = 0.;
    weight[i] = 0.;
  }
  for (i=0; i<3; i++){
    cd_i[i] = 0.;
    cd_o[i] = 0.;
    vi_o[i] = 0.;
    ve[i] = 0.;
    for (j=0; j<3; j++){
      t1[i][j] = t2[i][j] = t[i][j] =0.;
      t1_o[i][j] = t2_o[i][j] = t_o[i][j] = 0.;
      o_t[i][j] = 0.;
    }
  }
/* Initialisations done */

/* VIGNETTING FUNCTION read from FITS file (and smoothed) */
  if (plot_T && (int)(vigncode) == 1) {
    status = DALobjectOpen(vigndol, &vigPtr, status);
    if (status != ISDC_OK){
      RILlogMessage(NULL, Warning_2, "Cannot open object %s !", vigndol);
      return status;
    }
    else RILlogMessage(NULL, Log_2, "Use %s for vignetting in effective exposure time map", vigndol);

    status = DALarrayGetStruct(vigPtr, &DALtype, &nbAxes, axes, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot get structure for %s !", vigndol);
        return status;
      }
      else {
        for (i=0; i<nbAxes; i++) vignEnd[i] = axes[i];
        numE = (int)(vignEnd[2]);
/*        RILlogMessage(NULL, Log_3, "Nb of E- slices in the vignetting array is : %d", numE); */
      }
    if ( DALallocateDataBuffer((void**)&vslice, numE*sizeof(long),status) != ISDC_OK ) return status;
    for (k=0; k<numE; k++) vslice[k] = 1L;

    if ( DALallocateDataBuffer((void**)&vign3D, numValues*numValues*numE*sizeof(unsigned short),status) != ISDC_OK ) return status;
    arrayPtr = vign3D;
    status = DALarrayGetSection(vigPtr, nbAxes, vignStart, vignEnd, &DALtype, &nbValues, arrayPtr, status);
    if (status != ISDC_OK){
      RILlogMessage(NULL, Warning_2, "Cannot get section vign3D from %s", vigndol);
      return status;
    }
  } /* end if plot_T */


/***** READ HEADER ATTRIBUTES OF EACH INPUT IMAGE *****/
/* First build the selection string */

  strcpy(extSelect, "( (IMATYPE=='VARIANCE') || (IMATYPE=='VARIANCEU') )");
  sprintf(addSelect, " && ( (E_MIN >= %.3f) && (E_MAX <= %.3f) )", emin-prec_e, emax+prec_e);
  strcat(extSelect, addSelect);

  status = DAL3GENindexFindMember(group, imName, extSelect, &numHDU, &idxSel, status); /* List selected input images to get Ptr */
  if (ISDC_OK!=status) {
    RILlogMessage(NULL, Error_2, "Error %d finding VARIANCE images!", status);
    return status;
  }

  if (numHDU <= 0){
    RILlogMessage(NULL, Log_2, "*** No VARIANCE maps found with the given energies ! ***");
    RILlogMessage(NULL, Warning_3, "*** Cannot continue without VARIANCE maps... Stop!");
    return 1;
  }
  if (numImages != 0) num_E = numHDU/numImages;  /* Now we have num_E to keep getting all the necessary information */
  else{ 
    RILlogMessage(NULL, Warning_3, "Error determining nb of images or nb of E- bands... Stop!");
    return 1;
  }
  if (numHDU%numImages != 0){
    if (numImages > numHDU) RILlogMessage(NULL, Warning_1, "*** An E- band has fewer images than the nb of ScWs! ***");
    else RILlogMessage(NULL, Warning_2, "*** The total nb of images is not nb of ScWs X nb of E-bands as it should! ***");
    num_E = floor(numHDU/numImages +1);
  }

  if ( NULL == ( e_minT=malloc(num_E*numImages*sizeof(double)) ) ) return status;
  if ( NULL == ( e_min=(double **)malloc(num_E*sizeof(double *)) ) ) return status;
  for (i=0; i<num_E; i++) e_min[i] = &e_minT[i*numImages];
  if ( NULL == ( e_maxT=malloc(num_E*numImages*sizeof(double)) ) ) return status;
  if ( NULL == ( e_max=(double **)malloc(num_E*sizeof(double *)) ) ) return status;
  for (i=0; i<num_E; i++) e_max[i] = &e_maxT[i*numImages];
  if ( NULL == ( swidT=malloc(15*numImages*sizeof(char)) ) ) return status;
  if ( NULL == ( swid=(char **)malloc(numImages*sizeof(char *)) ) ) return status;
  for (i=0; i<numImages; i++){
    swid[i] = &swidT[i*15];
    for (j=0; j<num_E; j++) e_min[j][i] = e_max[j][i] = 0.;
  }
  if ( DALallocateDataBuffer((void**)&min_e, num_E*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&max_e, num_E*sizeof(double),status) != ISDC_OK ) return status;
  for (k=0; k<num_E; k++){
    min_e[k] = 80.;
    max_e[k] = 0.;
  }

  cnt = 0;
  for (overloop=1; overloop<=numHDU; overloop+=e_dol){  /* Loop over input Variance skymaps with the selected E-band(s) */
    loop = overloop-1-cnt;  /* "loop" registers the index number of ScWs */
    e_dol = 0;
    while (e_dol < num_E){
      
      /* SP: Different treatment if there's only one image */
      if (numHDU>1)status = DAL3GENindexGetMember(idxSel, imName, overloop+e_dol, &dolPtr, status); /* Open n-th (overloop) member of the list to get Ptr */
      else dolPtr=idxSel;
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get image nb %d! status: %d", overloop, status);
        return( status );
      }

    if (!comos){
      status = DALattributeGetChar(dolPtr, "SWID", swid[loop], NULL, NULL, status); /* Get SWID of the current loop */
      if (ISDC_OK!=status) {
        RILlogMessage(NULL, Error_1, "Error %d getting SWID!", status);
        return status;
      }
      if (loop == 0) RILlogMessage(NULL, Log_1, "The 1st SWID is : %s", swid[loop]);
      else if (loop == 1) RILlogMessage(NULL, Log_1, "The 2nd SWID is : %s", swid[loop]);
      else if (loop == 2) RILlogMessage(NULL, Log_1, "The 3rd SWID is : %s", swid[loop]);
      else RILlogMessage(NULL, Log_1, "The %dth SWID is : %s", loop+1, swid[loop]);

      status1=ISDC_OK;
      status1 = DALattributeGetReal(dolPtr, "EXPOSURE", &exp_t[loop], NULL, NULL, status1);
      if(status1 != ISDC_OK) status = DALattributeGetReal(dolPtr, "ONTIME", &exp_t[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute ONTIME nor EXPOSURE for image number %d", loop+1);
        return( status );
      }
      if (exp_t[loop] <= 0.) RILlogMessage(NULL, Log_3, "*** Warning: Image %d, EXPOSURE = %f ! ***", loop+1, exp_t[loop]);
    }

      status = DALattributeGetReal(dolPtr, "CD1_1", &cd1_1[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CD1_1 for image number %d", loop+1);
        return( status );
      }

      status = DALattributeGetReal(dolPtr, "CD2_1", &cd2_1[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CD2_1 for image number %d", loop+1);
        return( status );
      }

      status = DALattributeGetReal(dolPtr, "CD1_2", &cd1_2[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CD1_2 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "CD2_2", &cd2_2[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CD2_2 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "CRVAL1", &crval1[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CRVAL1 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "CRVAL2", &crval2[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CRVAL2 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "CRPIX1", &crpix1[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CRPIX1 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "CRPIX2", &crpix2[loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute CRPIX2 for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "E_MIN", &e_min[e_dol][loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute E_MIN for image number %d", loop+1);
        return( status );
      }
  
      status = DALattributeGetReal(dolPtr, "E_MAX", &e_max[e_dol][loop], NULL, NULL, status);
      if(status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot read attribute E_MAX for image number %d", loop+1);
        return( status );
      }

      if (plot_T && !comos){
        status = DALattributeGetInt(dolPtr, "RLIMIDX", &vslice[e_dol], NULL, NULL, status);
        if(status != ISDC_OK){
          RILlogMessage(NULL, Warning_1, "*** RLIMIDX not found -> Will chose default vignetting slice nb. 0 ***");
          vslice[e_dol] = 0L;
          status = ISDC_OK;
        }
      }

      if (e_min[e_dol][loop] < min_e[e_dol]) min_e[e_dol] = e_min[e_dol][loop];              /* Remember the actual */
      if (e_max[e_dol][loop] > max_e[e_dol]) max_e[e_dol] = e_max[e_dol][loop];              /*  Energy intervals. */
      e_dol++;
    } /* End of while (e_dol<num_E) loop */
    cnt += e_dol-1;

/* Initialise input WCS structure */
    cd_i[0] = cd1_1[loop]; cd_i[1] = cd1_2[loop]; cd_i[2] = cd2_1[loop]; cd_i[3] = cd2_2[loop];
    cdelt_i = fabs(cd2_2[loop]);
    wcs_inp = wcskinit((int)numValues, (int)numValues, ctype_1, ctype_2, crpix1[loop], crpix2[loop], crval1[loop], crval2[loop], cd_i, cdelt_i, cdelt_i, rot, 2000, 0);

  } /* End of "overloop" */
  if (numImages != loop+1) numImages = loop+1;
  RILlogMessage(NULL, Log_1, "*** Number of images to be combined %d ***", numImages);

  if (e_min != NULL) free(e_min);
  if (e_minT != NULL) free(e_minT);
  if (e_max != NULL) free(e_max);
  if (e_maxT != NULL) free(e_maxT);


/************************************* DEFINE THE OUTPUT MOSAIC SKYMAP *************************************/
/* v6.0.3: Prepare the output mosaic FITS file to include every energy band */
  if (strcmp(mosaic, "") ==0){
    if (comos) strcpy(fitstring, "jmx1_mosa_ima");
    else{
      switch (instrument) {
  	  case JMX1:
              strcpy(fitstring, "jmx1_mosa_ima");
  	      break;
  
  	  case JMX2:
              strcpy(fitstring, "jmx2_mosa_ima");
  	      break;  
  
  	  case JEMX:
              strcpy(fitstring, "jemx_mosa_ima");
  	      break;
  	     
  	  default:
	     RILlogMessage(NULL, Error_1, "INSTRUME(nt) %s is not supported! Please define the name of the output mosaic file", instName);
	     status=J_MOS_BAD_INSTRUMENT;
	     break;
      }
    }
  } else strcpy(fitstring, mosaic);

  status = DALobjectCreate(fitstring, DAL_DISK, template, &skyPtr, status);
  if (status != ISDC_OK){
    RILlogMessage(NULL, Warning_2, "Cannot create MOSAIC object, status = %d", status);
    return status;
  }

  crval1_min = 360.;
  crval1_max = 0.;
  crval2_min = 90.;
  crval2_max = -90.;
  for(loop=0; loop<numImages; loop++){
    if (crval1[loop] < crval1_min) crval1_min = crval1[loop];
    if (crval2[loop] < crval2_min) crval2_min = crval2[loop];
    if (crval1[loop] > crval1_max) crval1_max = crval1[loop];
    if (crval2[loop] > crval2_max) crval2_max = crval2[loop];
  }
  if ((crval1_max-crval1_min) >= 180.) crval1_max -= 360.;
  if (crval1_max < crval1_min){
    crval1_min = crval1_max;  
    for(loop=0; loop<numImages; loop++)
      if ((crval1[loop] > crval1_max) && crval1[loop] <= 180.) crval1_max = crval1[loop];
  }
  d_a = crval1_max-crval1_min;
  d_b = 90.-crval2_min;
  d_c = 90.-crval2_max;
  d_D = cos(d_b*dr)*cos(d_c*dr)+sin(d_b*dr)*sin(d_c*dr)*cos(d_a*dr);
  diamosaic = acos(d_D)/dr;
  if (diamout == 0.) diamout = diamosaic + 2.*radius;
  if ((diamout > 0.) && (diamosaic >= diamout)){
    RILlogMessage(NULL, Warning_1, "*** Total separation angle larger than %f degrees, some images may not be included ***", diamout);
    RILlogMessage(NULL, Log_2, "Largest separation angle is: %f degrees", diamosaic);
  }

  if (centerA < 0.){
    crval1_o = 0.5 * (crval1_min + crval1_max);
    if (crval1_o < 0.) crval1_o+= 360.;
    crval2_o = 0.5 * (crval2_min + crval2_max);
  }
  else {
    crval1_o = centerA;
    crval2_o = centerD;
  }
  RILlogMessage(NULL, Log_1, "Center of output mosaic skyimage: RA=%f, DEC=%f", crval1_o, crval2_o);

/* Define transformation from sky coord. to FITS */
  /*  cdelt = 0.03 ;  degrees per pixel*/
  if ((diamout > 0.) && (cdelt > 0.)){
    dimx = (int)(diamout / cdelt+.5);
    dimy = (int)(diamout / cdelt+.5);
  }
  if ( (dimx > (int)(sqrt(MOSAICSIZE)) ) || (dimy > (int)(sqrt(MOSAICSIZE)) ) ){
    RILlogMessage(NULL, Warning_3, "*** The total mosaic size is %d x %d pixels! ***", dimx,dimy);
    RILlogMessage(NULL, Warning_3, "*** This is larger than %d x %d pixels and the mosaic image may be corrupted ***", (int)(sqrt(MOSAICSIZE)),(int)(sqrt(MOSAICSIZE)));
  }
  if (radius >=0.) dimos = (int)((diamosaic + 2.*radius)/cdelt+.5);
  if ( (diamout < 0.) && (dimos > (int)(sqrt(MOSAICSIZE)) ) ){
    RILlogMessage(NULL, Warning_1, "*** The total mosaic size is %d x %d pixels! ***", dimos,dimos);
    RILlogMessage(NULL, Warning_1, "*** This is larger than %d x %d pixels available (MOSAICSIZE) so the mosaic image will be truncated ***", dimx,dimy);
  }

  crpix1_o = (float)(dimx)/2.-.5;
  crpix2_o = (float)(dimy)/2.-.5; 
  cd1_1_o = - cdelt;
  cd1_2_o = 0.0;
  cd2_1_o = 0.0;
  cd2_2_o = cdelt;
  ra_o = crval1_o;
  de_o = crval2_o;
  t1_o[0][0] = cos(ra_o*dr);
  t1_o[1][0] = sin(ra_o*dr);
  t1_o[2][0] = 0.;
  t1_o[0][1] = -sin(ra_o*dr);
  t1_o[1][1] = cos(ra_o*dr);
  t1_o[2][1] = 0.;
  t1_o[0][2] = 0.;
  t1_o[1][2] = 0.;
  t1_o[2][2] = 1.;
  t2_o[0][0] = cos(de_o*dr);
  t2_o[1][0] = 0.;
  t2_o[2][0] = sin(de_o*dr);
  t2_o[0][1] = 0.;
  t2_o[1][1] = 1.;
  t2_o[2][1] = 0.;
  t2_o[0][2] = -sin(de_o*dr);
  t2_o[1][2] = 0.;
  t2_o[2][2] = cos(de_o*dr);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++){
      t_o[i][j] = 0.;
      for (k=0; k<3; k++) t_o[i][j] += t1_o[i][k]*t2_o[k][j];
    }
  for (i=0; i<3; i++)
    for (j=0; j<3; j++) o_t[i][j] = t_o[j][i];

  cd_o[0] = cd1_1_o; cd_o[1] = cd1_2_o; cd_o[2] = cd2_1_o; cd_o[3] = cd2_2_o;
  wcs_out = wcskinit(dimx, dimy, ctype_1, ctype_2, crpix1_o, crpix2_o, crval1_o, crval2_o, cd_o, cdelt, cdelt, rot, 2000, 0);
  if (cdelt_i > 0.) pixfac = 1.08*(cdelt/cdelt_i)*(cdelt/cdelt_i);


/******************************** COMPOSE MOSAIC BY LOOPING OVER SKY IMAGES TO ADD FOR EACH ENERGY BAND ********************************/
/* First build the selection string according to user's input */
/******************************** V 6.0.2: THIS PART IS FOR BACKWARD COMPATIBILTY ONLY *****************************************************************/
if (strncmp(imSelect, "RAW_INT", 5)==0 || strncmp(imSelect, "RAW_REC", 5)==0 || strncmp(imSelect, "RES+SRC", 5)==0 || strcmp(imaType, "INTENSITY")==0){
  sprintf(intSelect, "(IMATYPE=='INTENSITY') && (E_MIN >= %.3f) && (E_MAX <= %.3f)", emin-prec_e, emax+prec_e);
  if (useA == 2) {
    sprintf(intSelect, "(IMATYPE=='RECONSTRUCTED') && (E_MIN >= %.3f) && (E_MAX <= %.3f)", emin-prec_e, emax+prec_e);
    sprintf(addSelect, " && (IMADESCR=='RES+SRC')");
  }
  else if (useA == 0) sprintf(addSelect, " && (IMADESCR=='RAW_INT')");   /* Make a restricted selection to find */
  else sprintf(addSelect, " && (IMADESCR=='RAW_RECT')");                 /*  the actual number of energy bands. */
} else
/*******************************************************************************************************************************************************/
{
  if (useA == 2) strcpy(intSelect, "( (IMATYPE=='RECONSTRUCTED') || (IMATYPE=='RECONSTRUCTEDU') )");
  else if (useA == 1) strcpy(intSelect, "( (IMATYPE=='RECTIFIED') || (IMATYPE=='RECTIFIEDU') )");
  else if (useA == 0) strcpy(intSelect, "( (IMATYPE=='RAWINTENSITY') || (IMATYPE=='RAWINTENSITYU') )");
  sprintf(addSelect, " && ( (E_MIN >= %.3f) && (E_MAX <= %.3f) )", emin-prec_e, emax+prec_e); /* Make a restricted selection to find */
  strcat (intSelect, addSelect);							      /*  the actual number of energy bands. */
}
  status = DAL3GENindexFindMember(group, imName, intSelect, &numIHDU, &intSel, status); /* List selected intensity maps to get Ptr */
  if (ISDC_OK!=status) {
    RILlogMessage(NULL, Error_2, "Error %d finding selected IMATYPE images!", status);
    return status;
  }

  sprintf(varSelect, "(IMATYPE=='VARIANCE') || (IMATYPE=='VARIANCEU') && (E_MIN >= %.3f) && (E_MAX <= %.3f)", emin-prec_e, emax+prec_e);
  status = DAL3GENindexFindMember(group, imName, varSelect, &numVHDU, &varSel, status); /* List selected variance maps to get Ptr */
  if (ISDC_OK!=status) {
    RILlogMessage(NULL, Error_2, "Error %d finding selected VARIANCE maps!", status);
    return status;
  }

  if (comos && plot_T){
    sprintf(expSelect, "(IMATYPE=='EXPOSURE') && (E_MIN >= %.3f) && (E_MAX <= %.3f)", emin-prec_e, emax+prec_e);
    status = DAL3GENindexFindMember(group, imName, expSelect, &numEHDU, &expSel, status); /* List selected Exposure maps to get Ptr */
    if (ISDC_OK!=status) {
      RILlogMessage(NULL, Error_2, "Error %d finding selected EXPOSURE maps!", status);
      return status;
    }
    if (numEHDU == 0) RILlogMessage(NULL, Log_0, "No Exposure maps in the input mosaics");

    sprintf(timSelect, "(IMATYPE=='OBS_TIME')");
    status = DAL3GENindexFindMember(group, imName, timSelect, &numTHDU, &timSel, status); /* List selected Exposure Time maps to get Ptr */
    if (ISDC_OK!=status) {
      RILlogMessage(NULL, Error_2, "Error %d finding selected OBS_TIME maps!", status);
      return status;
    }
    if (numTHDU == 0) RILlogMessage(NULL, Log_0, "No Exposure Time maps in the input mosaics");
  }

/*** Initialisations and memory allocations before main loop ***/
  if ( DALallocateDataBuffer((void**)&pvalue, numValues*numValues*sizeof(double),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&variance, numValues*numValues*sizeof(double),status) != ISDC_OK ) return status;
  if (comos){
    if ( DALallocateDataBuffer((void**)&expos, numValues*numValues*sizeof(double),status) != ISDC_OK ) return status;
    if ( DALallocateDataBuffer((void**)&time, numValues*numValues*sizeof(double),status) != ISDC_OK ) return status;
  }
  if ( NULL == ( pval_iT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
  if ( NULL == ( pval_i=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
  for (i=0; i<numValues; i++) pval_i[i] = &pval_iT[i*numValues];
  if ( NULL == ( var_iT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
  if ( NULL == ( var_i=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
  for (i=0; i<numValues; i++) var_i[i] = &var_iT[i*numValues];
  if ( NULL == ( vignT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
  if ( NULL == ( vign=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
  for (i=0; i<numValues; i++) vign[i] = &vignT[i*numValues];
  if (comos && plot_T){
    if ( NULL == ( exp_vT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
    if ( NULL == ( exp_v=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
    for (i=0; i<numValues; i++) exp_v[i] = &exp_vT[i*numValues];
    if ( NULL == ( exp_rT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
    if ( NULL == ( exp_r=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
    for (i=0; i<numValues; i++) exp_r[i] = &exp_rT[i*numValues];
  }
  if ( NULL == ( vigntmpT=malloc(numValues*numValues*sizeof(double)) ) ) return status;
  if ( NULL == ( vigntmp=(double **)malloc(numValues*sizeof(double *)) ) ) return status;
  for (i=0; i<numValues; i++) vigntmp[i] = &vigntmpT[i*numValues];
  if ( DALallocateDataBuffer((void**)&loopM, numImages*sizeof(int),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&rank, numImages*sizeof(int),status)!= ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&r_wei, numImages*sizeof(float),status) != ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&rwei, numImages*sizeof(float),status)!= ISDC_OK ) return status;
  if ( DALallocateDataBuffer((void**)&rweiMax, numImages*sizeof(float),status)!= ISDC_OK ) return status;
  if (!comos){
    if ( NULL == ( swid_redT=malloc(15*numImages*sizeof(char)) ) ) return status;
    if ( NULL == ( swid_red=(char **)malloc(numImages*sizeof(char *)) ) ) return status;
    for (i=0; i<numImages; i++) swid_red[i] = &swid_redT[i*15];
  }
  if (liScW){
    if ( NULL == ( scwT=malloc(DAL_MAX_ATT_COMMENT_SIZE*numImages*sizeof(char)) ) ) return status;
    if ( NULL == ( scw=(char **)malloc(numImages*sizeof(char *)) ) ) return status;
    for (i=0; i<numImages; i++) scw[i] = &scwT[i*DAL_MAX_ATT_COMMENT_SIZE];
  }

  for (i=0; i<xdim*ydim; i++){
    pvalue[i] = NaN;
    variance[i] = 0.;
    if (comos && plot_T){
      expos[i] = NaN;
      time[i] = NaN;
    }
  }
  for (i=0; i<xdim; i++)
    for (j=0; j<ydim; j++){
      pval_i[i][j] = NaN;
      if (comos && plot_T){
        exp_v[i][j] = NaN;
        exp_r[i][j] = NaN;
      }
      var_i[i][j] = 0.;
      vign[i][j] = 0.;
    }
  for (i=0; i<numImages; i++){
    loopM[i] =  numImages;
    rank[i] = 0;
    r_wei[i] = 0.;
  }


  for (e_dol=0; e_dol<num_E; e_dol++){   /*** Big loop over E- bands containing the mosaicking loop over the ScWs ***/
  /* Re-initialisation */
    cnt = 0;
    for (i=0; i<numImages; i++) rwei[i] = -1.;
    for (i=0; i<MOSAICSIZE; i++){
      pval_o[i] = 0.;
      var_o[i] = 0.;
      t_exp[i] = 0.;
      v_exp[i] = 0.;
      weight[i] = 0.;
    }

/*    sprintf(nrjSelect, "(E_MIN == %.3f) && (E_MAX == %.3f)", min_e[e_dol]-prec_e, max_e[e_dol]+prec_e);*/
    sprintf(nrjSelect, "((E_MIN>%.3f)&&(E_MIN<%.3f)) && ((E_MAX<%.3f)&&(E_MAX>%.3f))", min_e[e_dol]-prec_e,min_e[e_dol]+prec_e, max_e[e_dol]+prec_e,max_e[e_dol]-prec_e);
    if (numIHDU>1) status = DAL3GENindexFindMember(intSel, imName, nrjSelect, &numINT, &nrjSel, status); /* List selected input images to get Ptr */
    else {nrjSel=intSel; numINT=1;}  /* Different treatment if there's only one image */
    if (ISDC_OK!=status) {
      RILlogMessage(NULL, Error_2, "Error %d finding selected Intensity images for the given E- band!", status);
      return status;
    }
    if (numINT != numImages) {RILlogMessage(NULL, Warning_1, "*** Something is wrong: numINT= %d <> numIMages: %d! ***", numINT, numImages); cnt++;}

    if (numVHDU>1) status = DAL3GENindexFindMember(varSel, imName, nrjSelect, &numVAR, &nrjPtr, status); /* List selected input images to get Ptr */
    else {nrjPtr=varSel; numVAR=1;}  /* Different treatment if there's only one image */
    if (ISDC_OK!=status) {
      RILlogMessage(NULL, Error_2, "Error %d finding selected Variance images for the given E- band!", status);
      return status;
    }
    if (numINT != numVAR) {RILlogMessage(NULL, Warning_2, "*** Something is definitely wrong: numINT= %d <> numVAR= %d! ***", numINT, numVAR); cnt++;}

    if (comos && plot_T){
      if (numEHDU>1) status = DAL3GENindexFindMember(expSel, imName, nrjSelect, &numEXP, &expinp, status); /* List selected input mosaics to get Ptr */
      else if (numEHDU==1) {expinp=varSel; numEXP=1;}  /* Different treatment if there's only one image */
      if (ISDC_OK!=status) {
        RILlogMessage(NULL, Error_2, "Error %d finding selected Exposure input mosaics for the given E- band!", status);
        return status;
      }
      if (numEXP != numVAR) {RILlogMessage(NULL, Warning_2, "*** Something is wrong: numEXP= %d <> numVAR= %d! ***", numEXP, numVAR); cnt++;}
    }

    if (cnt > 0){
      numImages=numINT;
      if (cnt > 1) numImages=numVAR;
      RILlogMessage(NULL, Warning_3, "***** YOU HAVE BEEN WARNED: DO NOT TRUST THE RESULTS ! *****");
    }
    else RILlogMessage(NULL, Log_1, "Starting mosaic for E- band nb %d: %.2f - %.2f keV", e_dol, min_e[e_dol], max_e[e_dol]);

    /* Selecting the corresponding vignetting map */
    if (plot_T && !comos){
      slice = (int)(vslice[e_dol]);
      RILlogMessage(NULL, Log_2, " Using vignetting slice nb %d (counting from 0)", slice);
      for (p=0; p<xdim; p++)
        for (q=0; q<ydim; q++)
          vign[p][q] = (double)(*(vign3D+slice*xdim*ydim+q*xdim+p));

      /* Smoothing the vignetting map */
      for (i=1; i<xdim-1; i++)
        for (j=1; j<ydim-1; j++){
          vigntmp[i][j] = (vign[i-1][j-1]+vign[i][j-1]+vign[i+1][j-1] \
                        +  vign[i+1][j] + vign[i][j] + vign[i-1][j] \
                        +  vign[i-1][j+1]+vign[i][j+1]+vign[i+1][j+1])/9.;
          if (vigntmp[i][j] > maxvign) maxvign = vigntmp[i][j];  /* Prepare the normalisation */
        }
    }

    cnt = 0;
    for (loop=0; loop<numImages; loop++){  /* Loop over ScWs for input Intensity and Variance skymaps with the selected E-band(s) */
      if (comos) RILlogMessage(NULL, Log_1, "Acquiring mosaic nb %d:", loop+1);
      else{
        RILlogMessage(NULL, Log_1, "Acquiring ScW %d: %s", loop+1, swid[loop]);
        for (i=loop-cnt; i<numImages-cnt; i++) strcpy(swid_red[i], swid[i+cnt]);
      }
      if ( diamout > 0.){
        d_a = crval1[loop]-crval1_o;
        d_b = 90.-crval2[loop];
        d_c = 90.-crval2_o;
        d_D = cos(d_b*dr)*cos(d_c*dr)+sin(d_b*dr)*sin(d_c*dr)*cos(d_a*dr);
        if ( (acos(d_D)/dr-radius ) >= diamout/2. ){
          if (comos) RILlogMessage(NULL, Log_2, "* Excluding mosaic nb %d : outside the parameters-defined mosaic area *", loop+1);
          else RILlogMessage(NULL, Log_2, "* Excluding ScW %s : outside the parameters-defined mosaic area *", swid[loop]);
          cnt++;
          continue;
        }
      }

      for (i=0; i<MOSAICSIZE; i++) mark[i] = 0;

/*    Define transformation from FITS to sky coord. */

      cd_i[0] = cd1_1[loop]; cd_i[1] = cd1_2[loop]; cd_i[2] = cd2_1[loop]; cd_i[3] = cd2_2[loop];
      wcsreset(wcs_inp, crpix1[loop], crpix2[loop], crval1[loop], crval2[loop], cdelt_i, cdelt_i, rot, cd_i, 2000);

          ra_x = crval1[loop];
          de_x = crval2[loop];
          t1[0][0] = cos(ra_x*dr);
          t1[1][0] = sin(ra_x*dr);
          t1[2][0] = 0.;
          t1[0][1] = -sin(ra_x*dr);
          t1[1][1] = cos(ra_x*dr);
          t1[2][1] = 0.;
          t1[0][2] = 0.;
          t1[1][2] = 0.;
          t1[2][2] = 1.;
          t2[0][0] = cos(de_x*dr);
          t2[1][0] = 0.;
          t2[2][0] = sin(de_x*dr);
          t2[0][1] = 0.;
          t2[1][1] = 1.;
          t2[2][1] = 0.;
          t2[0][2] = -sin(de_x*dr);
          t2[1][2] = 0.;
          t2[2][2] = cos(de_x*dr);
          
          for (i=0; i<3; i++){
                for (j=0; j<3; j++){
                    t[i][j] = 0.;
                    for (k=0; k<3; k++) t[i][j] += t1[i][k]*t2[k][j];
                }
          }
  
  /*          Get skymaps inputs 
  */
	  /* SP: Different treatment if there's only one image */
	  if (numINT>1) status = DAL3GENindexGetMember(nrjSel, imName, loop+1, &intPtr, status); /* Open n-th (loop+1) ScW of the list to get Ptr */
	  else intPtr=nrjSel;

              if(status != ISDC_OK){
                RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get Intensity image nb %d! status: %d", loop+1, status);
                return( status );
              }

	      /* SP: Different treatment if there's only one image */
              if (numVAR>1) status = DAL3GENindexGetMember(nrjPtr, imName, loop+1, &varPtr, status); /* Open n-th (loop) ScW of the list to get Ptr */
	      else varPtr=nrjPtr;

              if(status != ISDC_OK){
                RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get Variance image nb %d! status: %d", loop+1, status);
                return( status );
              }

              if (comos && plot_T){
                if (numEXP>0) status = DAL3GENindexGetMember(expinp, imName, loop+1, &expPtr, status); /* Open n-th (loop) input mosaic of the list to get Ptr */
	        else RILlogMessage(NULL, Log_1, "No Exposure maps in the input mosaics -> no Exposure map in the output mosaic", status);
                if (numEXP ==1) expPtr=nrjPtr;
                if(status != ISDC_OK){
                  RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get Exposure map for the input mosaic nb %d! status: %d", loop+1, status);
                  return( status );
                }
                if (numTHDU>0) status = DAL3GENindexGetMember(timSel, imName, loop+1, &timPtr, status); /* Open n-th (loop) input mosaic of the list to get Ptr */
	        else RILlogMessage(NULL, Log_1, "No Exposure Time maps in the input mosaics -> no Exposure Time map in the output mosaic", status);
                if(status != ISDC_OK){
                  RILlogMessage(NULL, Warning_2,"DAL3GENindexGetMember cannot get Exposure Time map for the input mosaic nb %d! status: %d", loop+1, status);
                  return( status );
                }
              }

              if (!useA){  /* Check VIGNCOR */
                status = DALattributeGetInt(intPtr, "VIGNCOR", &vigncode, NULL, NULL, status);
                if(status != ISDC_OK){
                  if (chatter > 4) RILlogMessage(NULL, Warning_1, "*** VIGNCOR not found -> set to 0! ***");
                  vigncode = 0L;
                  status = ISDC_OK;
                }
                else if ( (int)(vigncode) == 1) RILlogMessage(NULL, Warning_2, "* Vignetting correction already applied in INTENSITY map! \
This is not consistent with the input image selected by the user!!*");

                status = DALattributeGetInt(varPtr, "VIGNCOR", &vigncode, NULL, NULL, status);
                if(status != ISDC_OK){
                  if (chatter > 4) RILlogMessage(NULL, Warning_1, "*** VIGNCOR not found -> set to 0! ***");
                  vigncode = 0L;
                  status = ISDC_OK;
                }
                else if ((int)(vigncode) != 0) RILlogMessage(NULL, Warning_2, "* Vignetting correction already applied in VARIANCE map! \
This is not consistent with the parameter mapSelect!! *");
              }

  
      /*      Read intensity maps 
      */
              arrayPtr = pvalue; DALtype = DAL_DOUBLE;
              status = DALarrayGetSection(intPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
              if (status != ISDC_OK){
                RILlogMessage(NULL, Warning_2, "Cannot get section pvalues (intensity) for image number %d", loop+1);
                return status;
              }
  
      /*      Read variance maps 
      */
              arrayPtr = variance;
              status = DALarrayGetSection(varPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
              if (status != ISDC_OK){
                RILlogMessage(NULL, Warning_2, "Cannot get section variance for image number %d", loop);
                return status;
              }
  
      /*      Read exposure maps in case of mosaicking of mosaics
      */
              if (comos && plot_T){
                arrayPtr = expos;
                status = DALarrayGetSection(expPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
                if (status != ISDC_OK){
                  RILlogMessage(NULL, Warning_2, "Cannot get section expos for input mosaic number %d", loop);
                  return status;
                }
                arrayPtr = time;
                status = DALarrayGetSection(timPtr, numAxes, startValues, endValues, &DALtype, &numValues, arrayPtr, status);
                if (status != ISDC_OK){
                  RILlogMessage(NULL, Warning_2, "Cannot get section time for input mosaic number %d", loop);
                  return status;
                }

              }

      /* Loop over individual pixels */
      for (p=0; p<xdim; p++){
        for (q=0; q<ydim; q++){
          index1 = q*xdim+p;
          if (plot_T && comos){
            exp_v[p][q] = isnan(expos[index1]) ? 0.0 : expos[index1];
            exp_r[p][q] = isnan(time[index1]) ? 0.0 : time[index1];
          }
          else if (plot_T && fabs(maxvign) > J_MOSA_EPSILON) vign[p][q] = vigntmp[p][q]/maxvign;
          pval_i[p][q] = isnan(pvalue[index1]) ? 0.0 : pvalue[index1]*pixfac;
          var_i[p][q] = isnan(variance[index1]) ? 0.0 : variance[index1]*pixfac/reduc;
/*          exp_i[p][q] = exp_t[loop]*reduc * vign[p][q];  Compute effective exposure with vignetting-function normalised to 1 */
          yz[0] = cd1_1[loop]*(p+1-crpix1[loop]) + cd1_2[loop]*(q+1-crpix2[loop]);
          yz[1] = cd2_1[loop]*(p+1-crpix1[loop]) + cd2_2[loop]*(q+1-crpix2[loop]);
          if ( (radius >=0.) && (sqrt(yz[0]*yz[0]+yz[1]*yz[1]) > radius) ) continue;

          /* Loop over pixels subdivisions  */          
          for (psub=-isub; psub<=isub; psub++){
          for (qsub=-isub; qsub<=isub; qsub++){
            dp = psub*dsub;
            dq = qsub*dsub;
            pix2wcs(wcs_inp, (double)(p+dp+1), (double)(q+dq+1), &Xpos, &Ypos);
            wcs2pix(wcs_out, Xpos, Ypos, &Xpix, &Ypix, &offscl);
            pix[0] = floor(Xpix+.5);
            pix[1] = floor(Ypix+.5);
            index2 = (pix[1]-1)*dimx+pix[0]-1;
            if ((pix[0]>=0) && (pix[0]<dimx) && (pix[1]>=0) && (pix[1]<dimy)){
/*if (index2 >0) printf("Xpix= %f  pix_0= %d    Ypix= %f  pix_1= %d      index2= %d\n", Xpix, pix[0], Ypix, pix[1], index2);*/
/*    COMPUTE WEIGTHS AND UPDATE MOSAIC SKYMAP */
              if (fabs(var_i[p][q]) > J_MOSA_EPSILON){
                wei = 1./var_i[p][q];
		weight[index2] += wei;
		pval_o[index2] += wei * pval_i[p][q];
		if (mark[index2]==0){
                  if (plot_T){
		    if (comos){
                      v_exp[index2] += exp_v[p][q];
                      t_exp[index2] += exp_r[p][q];
                    }
		    else{
                      v_exp[index2] += exp_t[loop] * vign[p][q]; /* Compute effective exposure times  v_exp[index2] += exp_t[loop]*reduc/pixfac * vign[p][q] */
                      t_exp[index2] += exp_t[loop]; /* ~ equivalent to compute raw exposure times by: t_exp[index2] += exp_t[loop]*reduc/pixfac */
                    }
                  } /* End of plot_T condition */
		  mark[index2] ++;
		  if ((index2 == dimx*(dimy/2-1)+dimx/2-1) && (centerA >=0.)) rwei[loop] = wei;
		} /* End of mark[index2]=0 condition */
 	      }  /* SP: Move end of 'fabs(var_i[p][q]) > J_MOSA_EPSILON' block here */
	    }   /* End of dimx,y>pix[0,1]>=0 condition */
	  }     /* End of qsub loop */
          }     /* End of psub loop */
        }       /* End of q loop */
      }         /* End of p loop */  
    }           /* End loop over every ScW (loop "loop") */

    for (index2=0; index2<MOSAICSIZE; index2++){
      /* SP: pixel must be divided by the sum of the individual weights */
      pval_o[index2] = (weight[index2]>0) ? pval_o[index2]/weight[index2] : NaN;
      /* SP: Var is 1/weight */
      var_o[index2] = (weight[index2]>0) ? 1/weight[index2] : NaN;
      if (!t_exp[index2]) t_exp[index2] = NaN;
      if (!v_exp[index2]) v_exp[index2] = NaN;
    }
  
    cntImages = numImages - cnt;
  
  /* List of relative weights produced on demand by the user into the logfile */
    if (centerA >=0.){
      RILlogMessage(NULL, Log_3, "");
      RILlogMessage(NULL, Log_3, "********************************************************************************");
      RILlogMessage(NULL, Log_3, "		       Ordered list of the input images ");
      RILlogMessage(NULL, Log_3, "                     for the E- band: %.2f keV - %.2f keV", min_e[e_dol], max_e[e_dol]);
      RILlogMessage(NULL, Log_3, "		    as a function of their relative weights ");
      RILlogMessage(NULL, Log_3, "	      at the given position: RA=%f  DEC=%f ", centerA, centerD);
      RILlogMessage(NULL, Log_3, "********************************************************************************");
      p = 0; k = 0; testM = 0;  /* Re-initialisation */
      for (i=0; i<cntImages; i++){
        loopM[i] = numImages+1;
        rank[i] = 0;
      }
      for (overloop=0; overloop<cntImages; overloop++){
        rweiMax[overloop] = 0.;
        if (rwei[overloop] < 0.){
          if (comos) RILlogMessage(NULL, Log_3, "The given position is outside the selected input radius in mosaic nb %d", overloop);
          else RILlogMessage(NULL, Log_3, "The given position is outside the selected input radius in ScW %s", swid_red[overloop]);
          p++;
        }
        for (loop=0; loop<cntImages; loop++){
          for(i=0; i<cntImages; i++) if (loopM[i] == loop) testM=1;
          if (testM){
            if (!rank[loop]) rank[loop] = overloop;
            testM = 0;
            continue;
          }
          if (rwei[loop] >= rweiMax[overloop]){
            rweiMax[overloop] = rwei[loop];
            k = loop;
          }
          if (rweiMax[0] > J_MOSA_EPSILON) r_wei[loop] = rwei[loop]/rweiMax[0];
        }
        loopM[overloop] = k;
      }
      RILlogMessage(NULL, Log_3, "");
      for (loop=0; loop<cntImages-p; loop++){
        if (!rank[loop]) rank[loop] = cntImages-p;
        if (comos) RILlogMessage(NULL, Log_3, "The input image from mosaic nb %d", loopM[loop]);
        else RILlogMessage(NULL, Log_3, "The input image from ScW %s", swid_red[loopM[loop]]);
        RILlogMessage(NULL, Log_3, "has rank %d with a relative weight equal to: %f", rank[loopM[loop]], r_wei[loopM[loop]]);
      }
      RILlogMessage(NULL, Log_3, "");
    }
 
  /* GENERATE OUTPUT MOSAIC FITS FILE */
    endMosaic[0] = dimx;
    endMosaic[1] = dimy;
  
    if (plot_I){
      nex++;
      RILlogMessage(NULL, Log_1, "-> Plot Intensity map (%.2f-%.2f keV) in %s.fits[%d]", min_e[e_dol], max_e[e_dol], fitstring, nex);
      status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &pvlPtr, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "DAL3GEN: Cannot create member for pval_o object");
        return status;
      }
      status = DALarrayModStruct( pvlPtr, DALdble, numAxes, endMosaic, status);
      if( status != ISDC_OK )  RILlogMessage(NULL, Warning_2, "DALarrayModStruct status : %d", status);
      if (!comos) status = DAL3GENattributeCopy(group, pvlPtr, "OGID,TSTART,TSTOP,TELAPSE", status);
      status = DAL3GENattributeCopy(intPtr, pvlPtr, "BUNIT,CHANMIN,CHANMAX,EVNTYPES,IMATYPE", status);
      status = DALattributePutChar( pvlPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
      status = DALattributePutChar( pvlPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
      status = DALattributePutReal( pvlPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
      status = DALattributePutReal( pvlPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
      status = DALattributePutReal( pvlPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( pvlPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( pvlPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( pvlPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( pvlPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
      status = DALattributePutReal( pvlPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
      status = DALattributePutReal( pvlPtr, "E_MIN", min_e[e_dol],"keV", "Lower Energy limit", status);
      status = DALattributePutReal( pvlPtr, "E_MEAN", (min_e[e_dol]+max_e[e_dol])/2.,"keV", "Mean Energy", status);
      status = DALattributePutReal( pvlPtr, "E_MAX", max_e[e_dol],"keV", "Upper Energy limit", status);
/*      status = DALattributePutChar( pvlPtr, "IMATYPE", "INTENSITY", " ", "The image shows intensity", status);*/
      /* Write list of input ScWs */
      if (liScW){
        status = DALattributePutChar(pvlPtr, "COMMENT", NULL, NULL, "List of input ScWs", status);
        for (loop=0; loop<cntImages; loop++){
          sprintf(scw[loop], "ScW_%d: ", loop+1);
          strcat(scw[loop], swid_red[loop]);
          status = DALattributePutChar(pvlPtr, NULL, NULL, NULL, scw[loop], status);
        }
      }
      arrayBuff = pval_o;
      status = DALarrayPutSection(pvlPtr, numAxes, startValues, endMosaic, DALdble, arrayBuff, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot put section pval_o object");
        return status;
      }
      /* SP: Stamp this object */
      status=CommonStampObject(pvlPtr,"This is the Intensity map",status);
      if (ISDC_OK!=status) { return(status); }
      DAL3GENindexUpdate(pvlPtr, skyPtr, status); status = 0;
    }
      
    if (plot_V){
      nex++;
      RILlogMessage(NULL, Log_1, "-> Plot Variance map (%.2f-%.2f keV) in %s.fits[%d]", min_e[e_dol], max_e[e_dol], fitstring, nex);
      status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &vmoPtr, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "DAL3GEN: Cannot create member for var_o object");
        return status;
      }
      status = DALarrayModStruct( vmoPtr, DALdble, numAxes, endMosaic, status);
      if( status != ISDC_OK )  RILlogMessage(NULL, Warning_2, "DALarrayModStruct status : %d", status);
      if (!comos) status = DAL3GENattributeCopy(group, vmoPtr, "OGID,TSTART,TSTOP,TELAPSE", status);
      status = DAL3GENattributeCopy(varPtr, vmoPtr, "BUNIT,CHANMIN,CHANMAX,EVNTYPES,IMATYPE", status);
      status = DALattributePutChar( vmoPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
      status = DALattributePutChar( vmoPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
      status = DALattributePutReal( vmoPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
      status = DALattributePutReal( vmoPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
      status = DALattributePutReal( vmoPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( vmoPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( vmoPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( vmoPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( vmoPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
      status = DALattributePutReal( vmoPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
      status = DALattributePutReal( vmoPtr, "E_MIN", min_e[e_dol],"keV", "Lower Energy limit", status);
      status = DALattributePutReal( vmoPtr, "E_MEAN", (min_e[e_dol]+max_e[e_dol])/2.,"keV", "Mean Energy", status);
      status = DALattributePutReal( vmoPtr, "E_MAX", max_e[e_dol],"keV", "Upper Energy limit", status);
/*      status = DALattributePutChar( vmoPtr, "IMATYPE", "VARIANCE", " ", "The image shows variance", status);*/
      /* Write list of input ScWs */
      if (liScW){
        status = DALattributePutChar(vmoPtr, "COMMENT", NULL, NULL, "List of input ScWs", status);
        for (loop=0; loop<cntImages; loop++){
          sprintf(scw[loop], "ScW_%d: ", loop+1);
          strcat(scw[loop], swid_red[loop]);
          status = DALattributePutChar(vmoPtr, NULL, NULL, NULL, scw[loop], status);
        }
      }
      arrayBuff = var_o;
      status = DALarrayPutSection(vmoPtr, numAxes, startValues, endMosaic, DALdble, arrayBuff, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot put section var_o object");
        return status;
      }
      /* SP: Stamp this object */
      status=CommonStampObject(vmoPtr,"This is the Variance map",status);
      if (ISDC_OK!=status) { return(status); }
      DAL3GENindexUpdate(vmoPtr, skyPtr, status);  status = 0;
    }
  
    for (index2=0; index2<MOSAICSIZE; index2++)
      if (var_o[index2] != 0.)
        pval_o[index2] = pval_o[index2]/sqrt(var_o[index2]);      /* Actual quantity to plot is significance: pv/sqrt(var) */ 

    if (plot_S){
      nex++;
      RILlogMessage(NULL, Log_1, "-> Plot Significance map (%.2f-%.2f keV) in %s.fits[%d]", min_e[e_dol], max_e[e_dol], fitstring, nex);
      status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &mapPtr, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "DAL3GEN: Cannot create member for new pval_o object");
        return status;
      }
      status = DALarrayModStruct( mapPtr, DALdble, numAxes, endMosaic, status);
      if( status != ISDC_OK )  RILlogMessage(NULL, Warning_2, "DALarrayModStruct status : %d", status);
      if (!comos) status = DAL3GENattributeCopy(group, mapPtr, "OGID,TSTART,TSTOP,TELAPSE", status);
      status = DAL3GENattributeCopy(intPtr, mapPtr, "CHANMIN,CHANMAX,EVNTYPES", status);
      status = DALattributePutChar( mapPtr, "BUNIT", "None", "","Unit for pixel values", status);
      status = DALattributePutChar( mapPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
      status = DALattributePutChar( mapPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
      status = DALattributePutReal( mapPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
      status = DALattributePutReal( mapPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
      status = DALattributePutReal( mapPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( mapPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( mapPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
      status = DALattributePutReal( mapPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
      status = DALattributePutReal( mapPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
      status = DALattributePutReal( mapPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
      status = DALattributePutReal( mapPtr, "E_MIN", min_e[e_dol],"keV", "Lower Energy limit", status);
      status = DALattributePutReal( mapPtr, "E_MEAN", (min_e[e_dol]+max_e[e_dol])/2.,"keV", "Mean Energy", status);
      status = DALattributePutReal( mapPtr, "E_MAX", max_e[e_dol],"keV", "Upper Energy limit", status);
      status = DALattributePutChar( mapPtr, "IMATYPE", "SIGNIFICANCE", " ", "The image shows significance", status);
      /* Write list of input ScWs */
      if (liScW){
        status = DALattributePutChar(mapPtr, "COMMENT", NULL, NULL, "List of input ScWs", status);
        for (loop=0; loop<cntImages; loop++){
          sprintf(scw[loop], "ScW_%d: ", loop+1);
          strcat(scw[loop], swid_red[loop]);
          status = DALattributePutChar(mapPtr, NULL, NULL, NULL, scw[loop], status);
        }
      }
      arrayBuff = pval_o;
      status = DALarrayPutSection(mapPtr, numAxes, startValues, endMosaic, DALdble, arrayBuff, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot put section new pval_o object");
        return status;
      }
      /* SP: Stamp this object */
      status=CommonStampObject(mapPtr,"This is the Significance map",status);
      if (ISDC_OK!=status) { return(status); }
      DAL3GENindexUpdate(mapPtr, skyPtr, status); status = 0;
    }

    if (plot_T){
      nex++;
      RILlogMessage(NULL, Log_1, "-> Plot effective Exposure time map in %s.fits[%d]", fitstring, nex);
      status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &extPtr, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "DAL3GEN: Cannot create member for v_exp object");
        return status;
      }
      status = DALarrayModStruct(extPtr, DALdble, numAxes, endMosaic, status);
      if( status != ISDC_OK )  RILlogMessage(NULL, Warning_2, "DALarrayModStruct status : %d", status);
      if (!comos) status = DAL3GENattributeCopy(group, extPtr, "OGID,TSTART,TSTOP,TELAPSE", status);
      status = DALattributePutChar(extPtr, "BUNIT", "second", "","Unit for pixel values", status);
      status = DALattributePutChar(extPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
      status = DALattributePutChar(extPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
      status = DALattributePutReal(extPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
      status = DALattributePutReal(extPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
      status = DALattributePutReal(extPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
      status = DALattributePutReal(extPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
      status = DALattributePutReal(extPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
      status = DALattributePutReal(extPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
      status = DALattributePutReal(extPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
      status = DALattributePutReal(extPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
      status = DALattributePutReal(extPtr, "E_MIN", min_e[e_dol],"keV", "Lower Energy limit", status);
      status = DALattributePutReal(extPtr, "E_MEAN", (min_e[e_dol]+max_e[e_dol])/2.,"keV", "Mean Energy", status);
      status = DALattributePutReal(extPtr, "E_MAX", max_e[e_dol],"keV", "Upper Energy limit", status);
      status = DALattributePutChar(extPtr, "IMATYPE", "EXPOSURE", "seconds", "The image shows effective Exposure times", status);
      /* Write list of input ScWs */
      if (liScW){
        status = DALattributePutChar(extPtr, "COMMENT", NULL, NULL, "List of input ScWs", status);
        for (loop=0; loop<cntImages; loop++){
          sprintf(scw[loop], "ScW_%d: ", loop+1);
          strcat(scw[loop], swid_red[loop]);
          status = DALattributePutChar(extPtr, NULL, NULL, NULL, scw[loop], status);
        }
      }
      arrayBuff = v_exp;
      status = DALarrayPutSection(extPtr, numAxes, startValues, endMosaic, DALdble, arrayBuff, status);
      if (status != ISDC_OK){
        RILlogMessage(NULL, Warning_2, "Cannot put section v_exp object");
        return status;
      }
      /* SP: Stamp this object */
      status=CommonStampObject(extPtr,"This is the effective Exposure time map",status);
      if (ISDC_OK!=status) { return(status); }
      DAL3GENindexUpdate(extPtr, skyPtr, status); status = 0;
    }

  }  /* End of big loop over E-bands */

  if (plot_T){
    RILlogMessage(NULL, Log_1, "-> Plot raw exposure Time map in %s.fits[%d]", fitstring, nex+1);
    status = DAL3GENindexCreateMember(skyPtr, NULL, NULL, &texPtr, status);
    if (status != ISDC_OK){
      RILlogMessage(NULL, Warning_2, "DAL3GEN: Cannot create member for t_exp object");
      return status;
    }
    status = DALarrayModStruct(texPtr, DALdble, numAxes, endMosaic, status);
    if( status != ISDC_OK )  RILlogMessage(NULL, Warning_2, "DALarrayModStruct status : %d", status);
    if (!comos) status = DAL3GENattributeCopy(group, texPtr, "OGID,TSTART,TSTOP,TELAPSE", status);
    status = DALattributePutChar(texPtr, "BUNIT", "second", "","Unit for pixel values", status);
    status = DALattributePutChar(texPtr, "CTYPE1", "RA---TAN","1962","Projection name", status);
    status = DALattributePutChar(texPtr, "CTYPE2", "DEC--TAN","1962 ","Projection name", status);
    status = DALattributePutReal(texPtr, "CRVAL1", crval1_o,"deg","Sky coord. ref. 1. axis", status);
    status = DALattributePutReal(texPtr, "CRVAL2", crval2_o,"deg","Sky coord. ref. 2. axis", status);
    status = DALattributePutReal(texPtr, "CD1_1", cd1_1_o,"deg", "Transformation", status);
    status = DALattributePutReal(texPtr, "CD2_2", cd2_2_o,"deg", "Transformation", status);
    status = DALattributePutReal(texPtr, "CD1_2", cd1_2_o,"deg", "Transformation", status);
    status = DALattributePutReal(texPtr, "CD2_1", cd2_1_o,"deg", "Transformation", status);
    status = DALattributePutReal(texPtr, "CRPIX1", crpix1_o,"pixel", "Sky pixel ref. 1. axis", status);
    status = DALattributePutReal(texPtr, "CRPIX2", crpix2_o,"pixel", "Sky pixel ref. 2. axis", status);
    status = DALattributePutChar(texPtr, "IMATYPE", "OBS_TIME", "seconds", "The image shows Observation Times", status);
    /* Write list of input ScWs */
    if (liScW){
      status = DALattributePutChar(texPtr, "COMMENT", NULL, NULL, "List of input ScWs", status);
      for (loop=0; loop<cntImages; loop++){
        sprintf(scw[loop], "ScW_%d: ", loop+1);
        strcat(scw[loop], swid_red[loop]);
        status = DALattributePutChar(texPtr, NULL, NULL, NULL, scw[loop], status);
      }
    }
    arrayBuff = t_exp;
    status = DALarrayPutSection(texPtr, numAxes, startValues, endMosaic, DALdble, arrayBuff, status);
    if (status != ISDC_OK){
      RILlogMessage(NULL, Warning_2, "Cannot put section t_exp object");
      return status;
    }
    /* SP: Stamp this object */
    status=CommonStampObject(texPtr,"This is the raw exposure Time map",status);
    if (ISDC_OK!=status) { return(status); }
    DAL3GENindexUpdate(texPtr, skyPtr, status); status = 0;
  }
    
  status = DALobjectClose(skyPtr, DAL_SAVE, status);
  if (status != ISDC_OK){
    RILlogMessage(NULL, Warning_2, "Cannot close object!");
    return status;
  }

  if (NULL!=imName)  DALfreeDataBuffer(imName,ISDC_OK);
  if (NULL!=exp_t)  DALfreeDataBuffer(exp_t,ISDC_OK);
  if (NULL!=crpix1)  DALfreeDataBuffer(crpix1,ISDC_OK);
  if (NULL!=crpix2)  DALfreeDataBuffer(crpix2,ISDC_OK);
  if (NULL!=cd1_1)  DALfreeDataBuffer(cd1_1,ISDC_OK);
  if (NULL!=cd2_1)  DALfreeDataBuffer(cd2_1,ISDC_OK);
  if (NULL!=cd1_2)  DALfreeDataBuffer(cd1_2,ISDC_OK);
  if (NULL!=cd2_2)  DALfreeDataBuffer(cd2_2,ISDC_OK);
  if (NULL!=crval1)  DALfreeDataBuffer(crval1,ISDC_OK);
  if (NULL!=crval2)  DALfreeDataBuffer(crval2,ISDC_OK);
  if (NULL!=loopM)  DALfreeDataBuffer(loopM,ISDC_OK);
  if (NULL!=rank)  DALfreeDataBuffer(rank,ISDC_OK);
  if (NULL!=r_wei)  DALfreeDataBuffer(r_wei,ISDC_OK);
  if (NULL!=rwei)  DALfreeDataBuffer(rwei,ISDC_OK);
  if (NULL!=rweiMax)  DALfreeDataBuffer(rweiMax,ISDC_OK);
  if (NULL!=swid) free(swid);
  if (NULL!=swid_red) free(swid_red);
  if ((liScW) && (NULL!=scw)) free(scw);
  if (plot_T){
    if (NULL!=vslice)  DALfreeDataBuffer(vslice,ISDC_OK);
    if (NULL!=vign3D)  DALfreeDataBuffer(vign3D,ISDC_OK);
  }
  if (comos && plot_T){
    if (NULL!=exp_v)  DALfreeDataBuffer(exp_v,ISDC_OK);
    if (NULL!=exp_r)  DALfreeDataBuffer(exp_r,ISDC_OK);
    if (NULL!=expos)  DALfreeDataBuffer(expos,ISDC_OK);
    if (NULL!=time)  DALfreeDataBuffer(time,ISDC_OK);
    if (NULL!=exp_vT) free(exp_vT);;
    if (NULL!=exp_rT) free(exp_rT);
  }
  wcsfree(wcs_inp);
  wcsfree(wcs_out);
RILclose(&fileRef);

  return 0;
}

/* *********************************************************************************************************** */
/*-------------------------------------------------------------------------------------------------------------*/
/* *********************************************************************************************************** */


/***************************************************************************

Function: instrumentGetCode

Description:
      
Utility function to get the Instrument code from a string with the name.

*****************************************************************************/

Instrument instrumentGetCode(const char* instrName){
  /* the following code does uppercase conversion, whitespace and punctuation
     compression to simplify comparison  */
  char convName[8];
  int  iCharOrig=0;
  int  iCharConv=0;
  
  while (instrName[iCharOrig]) { 
    if (isalpha(instrName[iCharOrig]) || isdigit(instrName[iCharOrig])) { 
      convName[iCharConv] = toupper(instrName[iCharOrig]);
      iCharConv++;
    }
    iCharOrig++;
  }
  convName[iCharConv] = '\0';
  
  if (!strcmp(convName,"JEMX"))    return JEMX;
  if (!strcmp(convName,"JMX"))     return JEMX;
  if (!strcmp(convName,"JEMX1"))   return JMX1;
  if (!strcmp(convName,"JMX1"))    return JMX1;
  if (!strcmp(convName,"JEMX2"))   return JMX2;
  if (!strcmp(convName,"JMX2"))    return JMX2;
  
  /* if we haven't returned yet, we have not recognized the instrument */
  return INST_NOT_RECOGNIZED;   
}
